/* 
	cif2ps.c

	Written 4/20/01 David_Harris@hmc.edu
	Updated 1/20/02 dh to speed up handling transparencies
	Ported to Linux 5/20/09 DMH 
	 - fixed size of token array to prevent crahses
   Update 22 April 2010 DMH
    - Support Cadence-style CIF by preprocessing with cadence_cif2ps

	This program generates postscript plots of an IC layout in CIF format.
	It handles color better than existing freely available
	postscript generators.  It does not handle arbitrary rotations.


   To produce very pretty diagrams for publication:
      use cif2ps with -nt no transparency option
      open .ps in Illustrator
      ungroup everything
      Select a metal3 wire.  
         Use Select * Same Fill Color to select all.  
         Group.
         Set Transparency to 70%.
         Use Windows * Pathfinder * Merge to merge all rectangles on the layer (seems very slow for M2 on a TinyChip)
         Set the outline color to a darker shade and an appropriate stroke width
     Repeat for Metal2, Metal1, Poly    
   

	Limitations:
	plots the last cell defined with DF, not the one given with C ; at the end of the file
	no way to change the plotter size right now
	the code to handle quad trees is rather messy now

	ideas:
		draw highlights around edges
		tokens is limited to 100.  Change to variable
		center port labels
		give options about aspect ratio / page size
		put date in postscript and on caption
		print layers on caption

	To do 1/28/03:
		draw outlines around edges
		handle black & white mode
		center labels
*/

/* #includes */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <string.h>
#include <ctype.h>

/* macros */

#define min(a,b) (((a) < (b)) ? (a) : (b))
#define max(a,b) (((a) > (b)) ? (a) : (b))

/* Constants */

#define MAXLAYERS 1000
#define MAXCOORDS 500
#define TREETHRESHOLD 500

/* Structures */

typedef struct polyL {
	int coords[MAXCOORDS]; /* not very memory efficient this way */
	int numcoords;
	int layer;
	struct polyL *next;
} polyL;

typedef struct point {
	int pos[2];
} point;

typedef struct outlineL {
	struct outlineL*next;
} outlineL;

typedef struct boxL {
	int pos[4];  // 0: dx, 1: dy, 2: left, 3: bot
	short layer;
	short visible;
	struct boxL *next;
} boxL;

typedef struct lineL {
	int x[2];
	int y;
	short visible;
	struct lineL *next;
} lineL;

typedef struct labelL {
	char label[80];
	int pos[2];
	struct labelL *next;
} labelL;

typedef struct cellL {
	int	cellNum;
	boxL *box;
	polyL *poly;
	labelL *label;
	struct cellL *next;
	struct cellInstL *inst;
} cellL;

typedef struct cellInstL {
	float transform[9];
	cellL *inst;
	struct cellInstL *next;
} cellInstL;

typedef struct box {
	int pos[4]; // 0: dx, 1: dy, 2: left, 3: bot
	short layer;
	short visible;
} box;

typedef struct boxquadtree {
	int numBoxes;
	box *boxes;
	int bounds[4]; // 0: dx, 1: dy, 2: left, 3: bot
	struct boxquadtree *tl;
	struct boxquadtree *tr;
	struct boxquadtree *bl;
	struct boxquadtree *br;
	struct boxquadtree *parent;
	int level;
} boxquadtree;

typedef struct layerMap {
	char name[80];
	int mix1, mix2; /* layer made from mix of previous layers, or -1 if not */
	double r, g, b;
	double opacity;
	int foreground;
} layerMap;

/* Globals */

int	cifBoundaries[4];
cellL *cells = NULL;
layerMap layers[MAXLAYERS];
boxL *flattenedbox[MAXLAYERS];
outlineL *flattenedoutline[MAXLAYERS];
boxquadtree *boxQuadTrees[MAXLAYERS];
polyL *flattenedpoly[MAXLAYERS];
lineL *horizsegs[MAXLAYERS];
char tokens[10000][80];
int numLayers;
int numBoxes = 0, numCells = 0, numPolys = 0, totalBoxes = 0, numInstances = 0;
int portlabel, caption, fontsize, notransparency;
char font[80], fname[80];

/* Function Prototypes */

void usage(char*);
void initICPlot(void);
void icplot(char*, char*);
void readLayerMap(char*);
void readCIF(char*);
void genOverlapShapesAfterFlattening(void);
void genOutlines(void);
lineL *genSegments(boxL *boxes, int dir);
int lineCompare(const void *a, const void *b);
int countBoxes(boxL *boxes);
outlineL *combineSegments(lineL *horiz, lineL *vert);
void checkOverlap(cellL*, boxL*, boxL*, int);
void coaf1(boxquadtree*, boxquadtree*, int);
void coaf2(boxquadtree*, boxquadtree*, int);
void coaf3(boxquadtree*, boxquadtree*, int);
void coaf4(boxquadtree*, boxquadtree*, int, int);
void checkOverlapAfterFlattening(boxquadtree*, boxquadtree*, int);
boxquadtree* makeBoxQuadTree(int);
void recursivelyMakeBoxQuadTree(boxquadtree*);
void splitTree(int, box*, boxquadtree*, int, int, int, int);
void flatten(void);
void recursiveFlatten(cellL*, float[]);
void writePS(char*, char*);
int searchLayerMap(char*);
cellL *findCell(int);
void mergeBoxes(void);
int mergeBoxPair(boxL*, boxL*);
boxL *copybox(boxL*, float[]);
polyL *copypoly(polyL*, float[]);
cellL* insertCell(cellL*, cellL*);
boxL* insertbox(boxL*, boxL*);
polyL* insertpoly(polyL*, polyL*);
labelL* insertlabel(labelL*, labelL*);
cellInstL* insertInst(cellInstL*, cellInstL*);
int tokenize(char*);
int linewrap(char*);
void matrixMul(float[], float[], float[]);
void newIdentityMatrix(float[]);
void transformBox(int*, int*, float[]);
void transformPoly(polyL*, polyL*, float[]);
void printStatistics(void);
void convPDF(char*);
void freeMem(void);
void freeBoxQuadTree(boxquadtree*);
void freeCell(cellL*);
void freeBox(boxL*);
void freePoly(polyL*);
void freeLabel(labelL*);
void freeCellInst(cellInstL*);
void freeLine(lineL*);

/* Main */

int main(int argc, char* argv[])
{
	char lname[80];
	int arg;

	portlabel = 1;
	caption = 1;
	strcpy(font, "Helvetica");
	fontsize = 12;
	strcpy(lname, "layermap.txt");

	if (argc < 2) {
	  usage(argv[0]);
	}
	else {
	  portlabel = 0; caption = 0; notransparency = 0;
		arg = 1;
		while (arg < argc-1) {
			if (strcmp(argv[arg], "-l") == 0) 
				strcpy(lname, argv[++arg]);
			else if (strcmp(argv[arg], "-p") == 0) { /* label ports */
				portlabel = 1; 
			}
			else if (strcmp(argv[arg], "-c") == 0) { /* include captions */
				caption = 1; 
			}
			else if (strcmp(argv[arg], "-nt") == 0) { /* don't generate transparency */
			  notransparency = 1;
			}
			else if (strcmp(argv[arg], "-f") == 0) { /* specify font */
				strcpy(font, argv[++arg]);
			}
			else if (strcmp(argv[arg], "-s") == 0) { /* specify font size */
				fontsize = atoi(argv[++arg]); 
			}
			else {
			  usage(argv[0]);
			}
			arg++;
		}
		strcpy(fname, argv[argc-1]);
	}
	icplot(lname, argv[0]);
}

void usage(char *fname)
{
  printf ("usage: %s [-l layermap.txt] [-p] [-c] [-nt] [-f font] [-s size] filename.cif\n  -p label top-level ports with font and size\n  -c print caption\n  -nt no transparency through layers (for CMOS VLSI Design figures)\n", fname);
  exit(1);
}


void icplot(char *layermap, char *progname)
{
	// this is the API for icplot
	// layermap: name of layermap file
	// progname: name of calling program (for header of postscript file)
	//   other values get passed as globals (file name, port labeling, ...)

	initICPlot();
	readLayerMap(layermap);
	readCIF(fname);
	mergeBoxes(); 
	flatten();
	genOverlapShapesAfterFlattening();
//	genOutlines();
	writePS(fname, progname);
	printStatistics();
	convPDF(fname);
	freeMem();
}

/* Functions */

void initICPlot(void)
{
	int i;

	cifBoundaries[0] = 0;
	cifBoundaries[1] = 0;
	cifBoundaries[2] = 0;
	cifBoundaries[3] = 0;

	for (i=0; i<MAXLAYERS; i++) {
		boxQuadTrees[i] = NULL;
	}
}

void readLayerMap(char *fname)
{
	FILE *layerfile;
	int i, curLayer, m1, m2;
	char line[1024];
	char str1[80], str2[80], str3[80], str4[80], str5[80], str6[80], str7[80];

	char fullfname[200];

	/* open the file */

	if ((layerfile = fopen(fname,"r")) == NULL) {
	  // can't find it in the local directory, look in install directory
	  strcpy(fullfname, "/cad/cif2ps/");
	  strcat(fullfname, fname);
	  fprintf(stderr, "can't find %s in local directory, looking for %s\n", 
		  fname, fullfname);
	  if ((layerfile = fopen(fullfname,"r")) == NULL) {
	    fprintf (stderr, "Unable to read layermap file %s\n", fullfname);
	    exit(1);
	  }
	}

	/* initialize layer table to be empty */
	for (i=0; i<MAXLAYERS; i++) 
		layers[i].name[0] = 0;
	numLayers = 0;

	/* read layer map */
	while (fgets(line, 1022, layerfile) != NULL) {
		if (line[0] == '#' || line[0] == '\n' || line[0] == '\r' || line[0] == 0) continue;
		str6[0] = 0; str7[0] = 0;
		sscanf(line,"%s %s %s %s %s %s %s %s",
			layers[numLayers].name, str1, str2, str3, str4, str5, 
			str6, str7);
		layers[numLayers].opacity = atof(str1);
		layers[numLayers].foreground = atoi(str2);
		layers[numLayers].r = atof(str3);
		layers[numLayers].g = atof(str4);
		layers[numLayers].b = atof(str5);
		m1 = searchLayerMap(str6);
		m2 = searchLayerMap(str7);
		if (m1 && m2) {
			layers[numLayers].mix1 = m1;
			layers[numLayers].mix2 = m2;
		}
		else {
			layers[numLayers].mix1 = -1;
			layers[numLayers].mix2 = -1;
		}
		if (layers[numLayers].opacity < 1 && !notransparency) {
			// create new layers to provide transparency
			curLayer = numLayers;
			for (i=0; i < curLayer; i++) {
				if (layers[i].foreground == 1) {
					layers[++numLayers].opacity = 1;
					strcpy(layers[numLayers].name, layers[i].name);
					strcat(layers[numLayers].name, layers[curLayer].name);
					layers[numLayers].foreground = 1;
					layers[numLayers].r = layers[curLayer].r*layers[curLayer].opacity +
						layers[i].r*(1-layers[curLayer].opacity);
					layers[numLayers].g = layers[curLayer].g*layers[curLayer].opacity +
						layers[i].g*(1-layers[curLayer].opacity);
					layers[numLayers].b = layers[curLayer].b*layers[curLayer].opacity +
						layers[i].b*(1-layers[curLayer].opacity);
					layers[numLayers].mix1 = i;
					layers[numLayers].mix2 = curLayer;
				}
			}
		}
		numLayers++;
		if (numLayers > MAXLAYERS) {
			fprintf(stderr, "Layer %d exceeds MAXLAYERS\n", numLayers);
			exit(1);
		}
	}
	fclose(layerfile);
	printf("Read layermap %s.  %d layers defined\n", fname, numLayers);
}

void readCIF(char *fname)
{
	FILE *infile;
	cellL *curCell;
	boxL *curbox;
	polyL *curpoly;
	labelL *curlabel;
	cellInstL *curInst;
	char line[1024], line2[1024], cmd[80], cmd2[80];
	int curLayer, cellNum;
	int t[2], x, y;
	int i, numtokens;
	float len, trans[9];
	int linewraps;
	int swap;
   int line2index, lineindex, linenum, pos, cmdpos;

	printf("reading CIF file %s...\n", fname);

	/* open the file */

	if ((infile = fopen(fname,"r")) == NULL) {
		fprintf (stderr, "Unable to read CIF file %s\n", fname);
		exit(1);
	}

	/* use a simple state machine to parse the CIF file */

	curLayer = 0;
	curCell = NULL; /* not in a cell yet */
   linenum = 0;
	while (fgets(line2, 1022, infile) != NULL) {
      line2index = 0;
      linenum++;
//      if (linenum > 5) exit(0);
 //     printf ("pt1\n");
      do {
//         printf("%d\n", linenum);
//         printf("considering %s\n", line2);
         lineindex = 0;
         while (line2[line2index] == ' '  || line2[line2index] == '\n' || line2[line2index] == '\r' || line2[line2index] == '\t') line2index++; // skip leading spaces
//         printf("pt2 %d\n", line2index);
         while (line2[line2index] != ';' && line2[line2index] != 0) {
            line[lineindex++] = line2[line2index++];
         } 
         if (line2[line2index] == ';') {
            line2index++; // skip trailing semicolon
            line[lineindex++] = ';';
         }
         line[lineindex] = 0;
//         if (linenum > 15460) printf("%d %d %d '%s'\n", linenum, lineindex, line2index, line);
 //        printf("%d%s\n", line);
		if (curCell == NULL) {
			if (line[0] == 'D' && line [1] == 'S') {
				/* found cell definition */
				curCell = (cellL*)malloc(sizeof(cellL));
				sscanf(line+2,"%d", &curCell->cellNum); // leave out scale factors
				curCell->box = NULL;
				curCell->poly = NULL;
				curCell->label = NULL;
				curCell->inst = NULL;
				curCell->next = NULL;
				numCells++;
			}
		}
		else {
 			sscanf(line,"%s", cmd);
         if (line[0] == 0) { 
            /* ignore blank line */
         } 
         else if (cmd[0] == 'D' && cmd[1] == 'F') {
				/* end of cell definition */
				cells = insertCell(curCell, cells);
				curCell = NULL;
			}
			else if (cmd[0] == 'L') {
				sscanf(line,"%s %s\n",cmd, cmd2);
				i = 0;
				do {
					if (cmd2[i] == ';') cmd2[i] = 0;
				} while (cmd2[i++] != 0);
				curLayer = searchLayerMap(cmd2);
			}
			else if (cmd[0] == 'B') {
				curbox = (boxL*)malloc(sizeof(boxL));
				curbox->layer = curLayer;
				curbox->next = NULL;
				curbox->visible = 1;
				numtokens = tokenize(line);
				// curbox->transform = identity;
				curbox->pos[0] = atoi(tokens[1]);
				curbox->pos[1] = atoi(tokens[2]);
				curbox->pos[2] = atoi(tokens[3]);
				curbox->pos[3] = atoi(tokens[4]);
				if (numtokens == 7) {
					/* rotation given, change coordinate system */
					x = atoi(tokens[5]);
					y = atoi(tokens[6]);
					if (x == 0) {
						swap = curbox->pos[1];
						curbox->pos[1] = curbox->pos[0];
						curbox->pos[0] = swap;
					}
					else if (y != 0) {
						printf("illegal rotation by %d %d\n", x, y);
					}
				}
				curbox->pos[2] -= curbox->pos[0]/2; // adjust center x to left edge;
				curbox->pos[3] -= curbox->pos[1]/2; // adjust center y to bottom edge
				curCell->box = insertbox(curbox,curCell->box);
			}
			else if (cmd[0] == 'P') {
				curpoly = (polyL*)malloc(sizeof(polyL));
				curpoly->layer = curLayer;
				curpoly->next = NULL;
				curpoly->numcoords = 0;
				i = 1;
            
				do {
				  numtokens = tokenize(line);
				  while (i+1 < numtokens) {
				    curpoly->coords[curpoly->numcoords++] = atoi(tokens[i]);
				    curpoly->coords[curpoly->numcoords++] = atoi(tokens[i+1]);
				    i = i+2;
				  }
				  linewraps = linewrap(line);
				  if (linewraps && fgets(line, 1022, infile) == NULL) linewraps = 0;
				  i = 0;
				} while (linewraps && curpoly->numcoords < MAXCOORDS);

				if (curpoly->numcoords == MAXCOORDS) {
					printf ("polygon too complex (more than %d points); trimming\n", MAXCOORDS/2);
				}
				curCell->poly = insertpoly(curpoly,curCell->poly);			
			}
			else if (cmd[0] == 'C') {
				// parse instantiation call 
				numtokens = tokenize(line);
            if (tokens[0][1] != 0) {
               // cell number is in same token as C
               i = 1;
               while (tokens[0][i] != 0) {
                  cmd[i-1] = tokens[0][i++];
               }
               cmd[i-1] = 0;
   				cellNum = atoi(cmd);
               i = 1;
            }
            else {
               // cell number in next token
   				cellNum = atoi(tokens[1]);
 				   i = 2; 
            } 
				curInst = (cellInstL*)malloc(sizeof(cellInstL));
				curInst->next = NULL;
				curInst->inst = findCell(cellNum);
				newIdentityMatrix(curInst->transform);
				while (i < numtokens) {
 					if (tokens[i][0] == 'T') {
                  if (tokens[i][1] != 0) { 
                     // cadence: T105,43125
                     pos = 1;
                     cmdpos = 0;
                     while (tokens[i][pos] != ',' && tokens[i][pos] != 0) {
                        cmd[cmdpos++] = tokens[i][pos++];
                     }
                     cmd[cmdpos] = 0;
                     t[0] = atoi(cmd);
                     /*if (tokens[i][pos] == 0) {
                        printf ("error parsing %s on %s\n", tokens[i], line);
                     }
                     else {
                        pos++;
                        cmdpos = 0;
                        while (tokens[i][pos] != 0) {
                           cmd[cmdpos++] = tokens[i][pos++];
                        }
                        cmd[cmdpos] = 0;
                        t[1] = atoi(cmd);
                     }
                     i = i+1;
                     */ 
                     t[1] = atoi(tokens[i+1]);
                     i = i+2;
                  } else {
                     // electric: T 105 43125
						   t[0] = atoi(tokens[i+1]);
						   t[1] = atoi(tokens[i+2]);
						   i = i+3;
                  }
						newIdentityMatrix(trans);
						trans[2*3+0] = (float)t[0];
						trans[2*3+1] = (float)t[1];
						matrixMul(curInst->transform, curInst->transform, trans);
					}
					else if (tokens[i][0] == 'R') {
						t[0] = atoi(tokens[i+1]);
						t[1] = atoi(tokens[i+2]);
						len = (float)sqrt((double)t[0]*t[0] + (float)t[1]*t[1]);
						newIdentityMatrix(trans);
						trans[0*3+0] = t[0]/len;
						trans[0*3+1] = t[1]/len;
						trans[1*3+0] = -t[1]/len;
						trans[1*3+1] = t[0]/len;
						matrixMul(curInst->transform, curInst->transform, trans);
						i = i+3;
					}
					else if (tokens[i][0] == 'M') {
						newIdentityMatrix(trans);
                  if (tokens[i][1] == 0) { 
						   if (tokens[i+1][0] == 'X') trans[0*3+0] = -1;
						   else if (tokens[i+1][0] == 'Y') trans[1*3+1] = -1;
						   else printf ("unreadable mirror statement:\n  %s\n", line);
						   i = i+2;
                  } else {
						   if (tokens[i][1] == 'X') trans[0*3+0] = -1;
						   else if (tokens[i][1] == 'Y') trans[1*3+1] = -1;
						   else printf ("unreadable mirror statement:\n  %s\n", line);
						   i = i+1;
                  }
						matrixMul(curInst->transform, curInst->transform, trans);
					}
               else {
                  printf("Unrecognized token %s in %s\n", tokens[i], line);
                  i = i+1;
               }
				}
				curCell->inst = insertInst(curInst, curCell->inst);

			}
			else if (cmd[0] == '9' && cmd[1] == '4') {
				numtokens = tokenize(line);
				if (numtokens != 4) printf("bad label syntax: %s\n", line);
				curlabel = (labelL*)malloc(sizeof(labelL));
				strcpy(curlabel->label, tokens[1]);
				curlabel->pos[0] = atoi(tokens[2]);
				curlabel->pos[1] = atoi(tokens[3]);
				curCell->label = insertlabel(curlabel, curCell->label);			
			}
         else if (cmd[0] == '9') {
           // skip comment line starting with 9
         }
         else {
            printf("Unrecognized line %s\n", line);
         }
		}
         while (line2[line2index] == ' '  || line2[line2index] == '\n' || line2[line2index] == '\r' || line2[line2index] == '\t') line2index++; // skip leading spaces
//     printf("%d %d '%c'\n", linenum, line2index, line2[line2index]);
      } while (line2[line2index] != 0);
	}
	fclose(infile);
      	printf("done reading cif\n");
}

void genOverlapShapesAfterFlattening(void)
{
	//boxL *g1, *g2;
	boxquadtree *q1, *q2;
	int i;

	/* traverse the list of boxes and create new layers where overlaps occur */

	printf ("generating overlap after flattening for %d layers...\n", numLayers);
	for (i=0; i<numLayers; i++) {
		printf(" %d", i+1);
		if (layers[i].mix1 != -1 && layers[i].mix2 != -1) {
			q1 = makeBoxQuadTree(layers[i].mix1);
			q2 = makeBoxQuadTree(layers[i].mix2);
			coaf1(q1, q2, i);
		}
	}
	printf("\n");
}

void genOutlines(void)
{
	int i;
	lineL *horiz, *vert;

	printf ("generating layer outlines...\n");
	for (i=0; i<numLayers; i++) {
		if (layers[i].mix1 == -1 && layers[i].mix2 == -1) {
			printf("%d (%s)  ", i+1, layers[i].name);
			horiz = genSegments(flattenedbox[i], 0);
			vert = genSegments(flattenedbox[i], 1);
			//eliminateSegments(horiz);
			//eliminateSegments(vert);
			flattenedoutline[i] = combineSegments(horiz, vert);
			horizsegs[i] = horiz;
		}
	}
	printf("\n");
}

lineL *genSegments(boxL *boxes, int dir)
{
	int numBoxes;
	int i;
	boxL *b;
	lineL *lines;
	lineL *segs, *s2;

	numBoxes = countBoxes(boxes);
	lines = (lineL*)malloc(2*numBoxes*sizeof(lineL));
	for (i=0, b=boxes; i<numBoxes; i++) {
		lines[2*i].visible = 1;
		lines[2*i].x[0] = b->pos[2+dir];
		lines[2*i].x[1] = b->pos[2+dir] + b->pos[0+dir];
		lines[2*i].y = b->pos[3-dir];
		lines[2*i+1].visible = 1;
		lines[2*i+1].x[0] = b->pos[2+dir];
		lines[2*i+1].x[1] = b->pos[2+dir] + b->pos[0+dir];
		lines[2*i+1].y = b->pos[3-dir] + b->pos[1-dir];
		b = b->next;
	}
	qsort(lines, 2*numBoxes, sizeof(lineL), lineCompare);
//	for (i=0; i<2*numBoxes; i++)
//		printf("\nseg %d: %d-%d, %d", i, lines[i].x[0], lines[i].x[1], lines[i].y);

	// eliminate segments (not implemented yet)
	// this may be difficult because boxes from different cells may overlap.
	// at this point I give up and drop it.  dh 1/31/03
	segs = NULL;
	for (i=0; i<2*numBoxes; i++) {
		s2 = (lineL*)malloc(sizeof(lineL));
		s2->x[0] = lines[i].x[0];
		s2->x[1] = lines[i].x[1];
		s2->y = lines[i].y;
		s2->visible = lines[i].visible;
		s2->next = segs;
		segs = s2;
	}

	// free array
	free(lines);
	
	return segs;
}

int lineCompare(const void *a, const void *b)
{
	return (((lineL*)a)->y - ((lineL*)b)->y);
}

int countBoxes(boxL *boxes)
{
	int i;
	boxL *b;

	b = boxes; i = 0;
	while (b != NULL) {
		i++;
		b = b->next;
	}

	return i;
}

outlineL *combineSegments(lineL *horiz, lineL *vert)
{
	return NULL;
}


boxquadtree *makeBoxQuadTree(int layer)
{
	boxL *g;
	int i, numBoxes = 0;

	// return if the quadtree is already generated
	if (boxQuadTrees[layer] != NULL) return boxQuadTrees[layer];

	// otherwise find number of boxes on this layer
	g = flattenedbox[layer];
	while (g != NULL) {
		if (g->visible) numBoxes++;
		g = g->next;
	}

	// and convert the list into a quad tree for faster processing

	// first allocate the quad tree
	boxQuadTrees[layer] = (boxquadtree*)malloc(sizeof(boxquadtree));
	boxQuadTrees[layer]->bounds[0] = cifBoundaries[0];
	boxQuadTrees[layer]->bounds[1] = cifBoundaries[1];
	boxQuadTrees[layer]->bounds[2] = cifBoundaries[2];
	boxQuadTrees[layer]->bounds[3] = cifBoundaries[3];
	boxQuadTrees[layer]->tl = NULL;
	boxQuadTrees[layer]->tr = NULL;
	boxQuadTrees[layer]->bl = NULL;
	boxQuadTrees[layer]->br = NULL;
	boxQuadTrees[layer]->numBoxes = numBoxes;
	boxQuadTrees[layer]->boxes = (box*)malloc(numBoxes*sizeof(box));
	boxQuadTrees[layer]->parent = NULL;
	boxQuadTrees[layer]->level = 0;

	// then copy the boxes into the tree
	g = flattenedbox[layer];
	i = 0;
	while (g != NULL) {
		if (g->visible) {
			boxQuadTrees[layer]->boxes[i].pos[0] = g->pos[0];
			boxQuadTrees[layer]->boxes[i].pos[1] = g->pos[1];
			boxQuadTrees[layer]->boxes[i].pos[2] = g->pos[2];
			boxQuadTrees[layer]->boxes[i].pos[3] = g->pos[3];
			boxQuadTrees[layer]->boxes[i].layer = g->layer;
			boxQuadTrees[layer]->boxes[i].visible = 1;
			i++;
		}
		g = g->next;
	}

	// if there are too many boxes in this layer of the tree, create subtrees
	if (numBoxes > TREETHRESHOLD) 
		recursivelyMakeBoxQuadTree(boxQuadTrees[layer]);
	
	return boxQuadTrees[layer];
}

void recursivelyMakeBoxQuadTree(boxquadtree *q)
{
	int i, numBoxes;

	q->tl = (boxquadtree*)malloc(sizeof(boxquadtree)); q->tl->parent = q; q->tl->level = q->level*10+1;
	q->tr = (boxquadtree*)malloc(sizeof(boxquadtree)); q->tr->parent = q; q->tr->level = q->level*10+2;
	q->bl = (boxquadtree*)malloc(sizeof(boxquadtree)); q->bl->parent = q; q->bl->level = q->level*10+3;
	q->br = (boxquadtree*)malloc(sizeof(boxquadtree)); q->br->parent = q; q->br->level = q->level*10+4;

	// split boxes into subtrees where they fit
	splitTree(q->numBoxes, q->boxes, q->bl, q->bounds[0], q->bounds[1], (q->bounds[0]+q->bounds[2])/2, (q->bounds[1]+q->bounds[3])/2);
	splitTree(q->numBoxes, q->boxes, q->br, (q->bounds[0]+q->bounds[2])/2, q->bounds[1], q->bounds[2], (q->bounds[1]+q->bounds[3])/2);
	splitTree(q->numBoxes, q->boxes, q->tl, q->bounds[0], (q->bounds[1]+q->bounds[3])/2, (q->bounds[0]+q->bounds[2])/2, q->bounds[3]);
	splitTree(q->numBoxes, q->boxes, q->tr, (q->bounds[0]+q->bounds[2])/2, (q->bounds[1]+q->bounds[3])/2, q->bounds[2], q->bounds[3]);

	// and leave boxes that span the subtrees at the top level
	numBoxes = 0;
	for (i=0; i<q->numBoxes; i++) {
		if (q->boxes[i].visible) {
			// q->boxes[numBoxes] = q->boxes[i];
			q->boxes[numBoxes].layer = q->boxes[i].layer;
			q->boxes[numBoxes].visible = 1;
			q->boxes[numBoxes].pos[0] = q->boxes[i].pos[0];
			q->boxes[numBoxes].pos[1] = q->boxes[i].pos[1];
			q->boxes[numBoxes].pos[2] = q->boxes[i].pos[2];
			q->boxes[numBoxes].pos[3] = q->boxes[i].pos[3];
			numBoxes++;
		}
	}

	q->numBoxes = numBoxes;
}

void splitTree(int numBoxes, box *boxes, boxquadtree *q, int left, int bot, int right, int top)
{
	int i;
	int count = 0;

	// count how many boxes are in subtree
	for (i = 0; i<numBoxes; i++) {
		if (boxes[i].visible && boxes[i].pos[2] >= left && boxes[i].pos[3] >= bot && 
			(boxes[i].pos[2]+boxes[i].pos[0]) <= right && (boxes[i].pos[3]+boxes[i].pos[1]) <= top) count++;
	}
	// and copy them into a new array for the subtree
	q->boxes = (box*)malloc(count*sizeof(box));
	count = 0;
	for (i=0; i<numBoxes; i++) {
		if (boxes[i].visible && boxes[i].pos[2] >= left && boxes[i].pos[3] >= bot && 
			(boxes[i].pos[2]+boxes[i].pos[0]) <= right && (boxes[i].pos[3]+boxes[i].pos[1]) <= top) {
			//q->boxes[count] = boxes[i];
			q->boxes[count].layer = boxes[i].layer;
			q->boxes[count].visible = 1;
			q->boxes[count].pos[0] = boxes[i].pos[0];
			q->boxes[count].pos[1] = boxes[i].pos[1];
			q->boxes[count].pos[2] = boxes[i].pos[2];
			q->boxes[count].pos[3] = boxes[i].pos[3];
			boxes[i].visible = 0; // mark box for removal from upper level
			count++;
		}
	}
		
	q->numBoxes = count;
	q->bounds[0] = left;
	q->bounds[1] = bot;
	q->bounds[2] = right;
	q->bounds[3] = top;
	q->tl = NULL;
	q->tr = NULL;
	q->bl = NULL;
	q->br = NULL;
	
	if (count > TREETHRESHOLD) {
		recursivelyMakeBoxQuadTree(q);
	}
}

void mergeBoxes(void)
{
	cellL *c;
	boxL *b1, *b2;
	int changed;
	int numMerged = 0;

	printf("merging boxes for %d cells...\n", numCells);
	c = cells;
	while (c != NULL) {
		do {
			changed = 0;
			b1 = c->box;
			while (b1 != NULL) {
				b2 = b1->next;
				while (b2 != NULL) {
					if (b1->layer == b2->layer && b1->visible && b2->visible) 
						if (mergeBoxPair(b1, b2)) changed = 1;
					b2 = b2->next;
				}
				b1 = b1->next;
			}
		} while (changed);
		c = c->next;
		numMerged++;
	}
	printf("\n");
}

int mergeBoxPair(boxL *b1, boxL *b2)
{
	int t[3], b[3], l[3], r[3];

	t[0] = b1->pos[3] + b1->pos[1];
	b[0] = b1->pos[3];
	l[0] = b1->pos[2];
	r[0] = b1->pos[2] + b1->pos[0];
	t[1] = b2->pos[3] + b2->pos[1];
	b[1] = b2->pos[3];
	l[1] = b2->pos[2];
	r[1] = b2->pos[2] + b2->pos[0];

	// if the boxes coincide, expand the first one and hide the second one

	if (t[0] == t[1] && b[0] == b[1] && (min(r[0], r[1]) > max(l[0], l[1]))) {
		l[2] = min(l[0], l[1]);
		r[2] = max(r[0], r[1]);
		b1->pos[0] = (r[2]-l[2]);
		b1->pos[1] = (t[0]-b[0]);
		b1->pos[2] = (l[2]);
		b1->pos[3] = (b[0]);
		b2->visible = 0;
		return 1;
	}
	else if (r[0] == r[1] && l[0] == l[1] && (min(t[0], t[1]) > max(b[0], b[1]))) {
		b[2] = min(b[0], b[1]);
		t[2] = max(t[0], t[1]);
		b1->pos[0] = (r[0]-l[0]);
		b1->pos[1] = (t[2]-b[2]);
		b1->pos[2] = (l[0]);
		b1->pos[3] = (b[2]);
		b2->visible = 0;
		return 1;
	}
	// if one completely covers another, hide the covered box
	else if (r[0] >= r[1] && l[0] <= l[1] && t[0] >= t[1] && b[0] <= b[1]) {
		b2->visible = 0;
		return 1;
	}
	else if (r[1] >= r[0] && l[1] <= l[0] && t[1] >= t[0] && b[1] <= b[0]) {
		b1->visible = 0;
		return 1;
	}
	return 0;
}

void checkOverlap(cellL *c, boxL *b1, boxL *b2, int layerNum)
{
	int t[3], b[3], l[3], r[3];
	boxL *curbox;

	t[0] = b1->pos[3] + b1->pos[1];
	b[0] = b1->pos[3];
	l[0] = b1->pos[2];
	r[0] = b1->pos[2] + b1->pos[0];
	t[1] = b2->pos[3] + b2->pos[1];
	b[1] = b2->pos[3];
	l[1] = b2->pos[2];
	r[1] = b2->pos[2] + b2->pos[0];

	t[2] = t[0] < t[1] ? t[0] : t[1];
	b[2] = b[0] > b[1] ? b[0] : b[1];
	l[2] = l[0] > l[1] ? l[0] : l[1];
	r[2] = r[0] < r[1] ? r[0] : r[1];

	if (t[2] > b[2] && r[2] > l[2]) {
		/* create overlap layer */
		curbox = (boxL*)malloc(sizeof(boxL));
		curbox->layer = layerNum;
		curbox->next = NULL;
		curbox->pos[0] = (r[2]-l[2]);
		curbox->pos[1] = (t[2]-b[2]);
		curbox->pos[2] = (l[2]);
		curbox->pos[3] = (b[2]);
		curbox->visible = 1;
		if (t[2] == t[0] && b[2] == b[0] && l[2] == l[0] && r[2] == r[0]) 
			b1->visible = 0;
		if (t[2] == t[1] && b[2] == b[1] && l[2] == l[1] && r[2] == r[1]) 
			b2->visible = 0;
		c->box = insertbox(curbox, c->box);
	}
}

void coaf1(boxquadtree *q1, boxquadtree *q2, int layernum)
{
	coaf2(q1, q2, layernum);
	if (q1->tl != NULL) {
		coaf3(q1->tl, q2, layernum);
		coaf3(q1->tr, q2, layernum);
		coaf3(q1->bl, q2, layernum);
		coaf3(q1->br, q2, layernum);
		if (q2->tl != NULL) {
			coaf1(q1->tl, q2->tl, layernum);
			coaf1(q1->tr, q2->tr, layernum);
			coaf1(q1->bl, q2->bl, layernum);
			coaf1(q1->br, q2->br, layernum);
		}
		else {
			coaf4(q1->tl, q2, layernum, 0);
			coaf4(q1->tr, q2, layernum, 0);
			coaf4(q1->bl, q2, layernum, 0);
			coaf4(q1->br, q2, layernum, 0);
		}
	}
}

void coaf2(boxquadtree *q1, boxquadtree *q2, int layernum)
{
	checkOverlapAfterFlattening(q1, q2, layernum);
	if (q2->tl != NULL) {
		coaf2(q1, q2->tl, layernum);
		coaf2(q1, q2->tr, layernum);
		coaf2(q1, q2->bl, layernum);
		coaf2(q1, q2->br, layernum);
	}
}

void coaf3(boxquadtree *q1, boxquadtree *q2, int layernum)
{
	checkOverlapAfterFlattening(q1, q2, layernum);
	if (q2->parent != NULL) 
		coaf3(q1, q2->parent, layernum);
}

void coaf4(boxquadtree *q1, boxquadtree *q2, int layernum, int check)
{
	if (check) {
		coaf3(q1, q2, layernum);
	}
	if (q1->tl != NULL) {
		coaf4(q1->tl, q2, layernum, 1);
		coaf4(q1->tr, q2, layernum, 1);
		coaf4(q1->bl, q2, layernum, 1);
		coaf4(q1->br, q2, layernum, 1);
	}
}

void checkOverlapAfterFlattening(boxquadtree *q1, boxquadtree *q2, int layerNum)
{
	int j, k;
	int t[3], b[3], l[3], r[3];
	boxL *curbox;

	// check overlap of boxes at this level of the quad tree
	if (q1->numBoxes && q2->numBoxes) {
		for (j=0; j<q1->numBoxes; j++) {
			t[0] = q1->boxes[j].pos[3] + q1->boxes[j].pos[1];
			b[0] = q1->boxes[j].pos[3];
			l[0] = q1->boxes[j].pos[2];
			r[0] = q1->boxes[j].pos[2] + q1->boxes[j].pos[0];

			for (k=0; k < q2->numBoxes; k++) {
				t[1] = q2->boxes[k].pos[3] + q2->boxes[k].pos[1];
				b[1] = q2->boxes[k].pos[3];
				l[1] = q2->boxes[k].pos[2];
				r[1] = q2->boxes[k].pos[2] + q2->boxes[k].pos[0];

				t[2] = t[0] < t[1] ? t[0] : t[1];
				b[2] = b[0] > b[1] ? b[0] : b[1];
				l[2] = l[0] > l[1] ? l[0] : l[1];
				r[2] = r[0] < r[1] ? r[0] : r[1];

				if (t[2] > b[2] && r[2] > l[2]) {
					/* create overlap layer */
					curbox = (boxL*)malloc(sizeof(boxL));
					curbox->layer = layerNum;
					curbox->next = NULL;
					curbox->pos[0] = (r[2]-l[2]);
					curbox->pos[1] = (t[2]-b[2]);
					curbox->pos[2] = (l[2]);
					curbox->pos[3] = (b[2]);
					curbox->visible = 1;
					if (t[2] == t[0] && b[2] == b[0] && l[2] == l[0] && r[2] == r[0]) q1->boxes[j].visible = 0;
					if (t[2] == t[1] && b[2] == b[1] && l[2] == l[1] && r[2] == r[1]) q2->boxes[k].visible = 0;
					flattenedbox[layerNum] = insertbox(curbox, flattenedbox[layerNum]);
				}
			}
		}
	}
}

void flatten(void)
{
	float ident[9];
	int i;
	
	newIdentityMatrix(ident);
	for (i=0; i<numLayers; i++) {
		flattenedbox[i] = NULL;
	}

	/* for now, assume last cell is top level.  Change this to recognize C *** at top of CIF */

	printf("flattening...\n");
	recursiveFlatten(cells, ident);
}

void recursiveFlatten(cellL *top, float m[9])
{
	cellInstL *inst;
	boxL *box, *newbox;
	polyL *poly, *newpoly;
	float tm[9];

	/* add boxes from this cell */
	box = top->box;
	while (box != NULL) {
		if (box->visible) {
			newbox = copybox(box, m);
			flattenedbox[newbox->layer] = insertbox(newbox, flattenedbox[newbox->layer]);
		}
		box = box->next;
	}

	/* add polygons from this cell */
	poly = top->poly;
	while (poly != NULL) {
		newpoly = copypoly(poly, m);
		flattenedpoly[newpoly->layer] = insertpoly(newpoly, flattenedpoly[newpoly->layer]);
		poly = poly->next;
	}

	/* recursively traverse subinstances */
	inst = top->inst;
	while (inst != NULL) {
		numInstances++;
		matrixMul(tm, inst->transform, m);
		recursiveFlatten(inst->inst, tm);
		inst = inst->next;
	}
}

void writePS(char *fname, char *progname)
{
	FILE *psfile;
	int i, j;
	polyL *p;
	boxL *g;
	labelL *l;
	float scale;
	char *date="4/21/01"; /* *** change this to look up date */
	//	float size = 8;
	int border = 20;
	int pageHeight = 8*72, pageWidth = 8*72;
	int w, h;
	int offset;

	printf("Writing postscript file %s.ps...\n", fname);
	strcat(fname, ".ps");

	if ((psfile = fopen(fname,"w")) == NULL) {
		fprintf (stderr, "Unable to write %s\n", fname);
		exit(1);
	}
 
	/* Header info */\
	fprintf(psfile, "%%!PS-Adobe-1.0\n");
	fprintf(psfile, "%% Created %s by %s (written by David_Harris@hmc.edu)\n\n", date, progname);

	/* Coordinate system */
	offset = (cifBoundaries[3]-cifBoundaries[1])/10;
	if (caption) cifBoundaries[1] -= offset;
	fprintf(psfile, "%% Min X: %d  min Y: %d  max X: %d  max Y: %d\n", 
		cifBoundaries[0], cifBoundaries[1], cifBoundaries[2], cifBoundaries[3]);
	scale = pageWidth/(float)(cifBoundaries[2]-cifBoundaries[0]);
	if (pageHeight/(float)(cifBoundaries[3]-cifBoundaries[1]) < scale)
		scale = pageHeight/(float)(cifBoundaries[3]-cifBoundaries[1]); 
	fprintf(psfile, "%f %f scale\n", scale, scale);
	fprintf(psfile, "%d neg %d neg translate\n", cifBoundaries[0]-border, cifBoundaries[1]-border);

	/* set font */
	fprintf(psfile, "/%s findfont %f scalefont setfont\n", font, fontsize/scale);

	/* print caption if applicable */
	if (caption) {
		fprintf(psfile, "\n%d %d moveto\n", cifBoundaries[0], cifBoundaries[1]+offset/2);
		fprintf(psfile, "(%s) show\n", fname);
	}

	/* define box command to make rectangles more memory efficient */
	fprintf(psfile, "\n/bx \n  { /h exch def /w exch def /x exch def /y exch def\n");
	fprintf(psfile, "    newpath x y moveto w 0 rlineto 0 h rlineto w neg 0 rlineto closepath fill } def\n"); 

	/* draw layers */

	for (i=0; i<numLayers; i++) {
		// skip drawing layers that are white
		if (layers[i].r != 1 || layers[i].g != 1 || layers[i].b != 1) {
			g = flattenedbox[i];
			p = flattenedpoly[i];
			if (g != NULL || p != NULL) {
				fprintf(psfile,"\n%% Layer %d %s", i, layers[i].name);
				if (layers[i].mix1 != -1) fprintf(psfile," mix of %s and %s", layers[layers[i].mix1].name, layers[layers[i].mix2].name);
				fprintf(psfile,"\n%f %f %f setrgbcolor\n", layers[i].r, layers[i].g, layers[i].b);
			}
			while (g != NULL) {
				if (g->visible) {
					w = g->pos[0];
					h = g->pos[1];
					fprintf(psfile, "%d %d %d %d bx\n",	g->pos[3], g->pos[2], w, h);	  
					numBoxes++;
				}
				g = g->next;
				totalBoxes++;
			}
			while (p != NULL) {
				fprintf(psfile, "newpath %d %d moveto \n", p->coords[0], p->coords[1]);
				for (j=2; j<p->numcoords; j+=2) {
					fprintf(psfile, "        %d %d lineto \n", p->coords[j], p->coords[j+1]);
				}
				fprintf(psfile, "closepath %s\n", i==0 ? "stroke" : "fill");
				p = p->next;
				numPolys++;
			}
/*			lineL *h = horizsegs[i];
			fprintf(psfile, "\n0 0 0 setrgbcolor\n newpath\n");
			while (h != NULL) {
				fprintf(psfile, "%d %d moveto %d %d lineto\n", h->x[0], h->y, h->x[1], h->y);
				h = h->next;
			}
			fprintf(psfile, "closepath stroke\n");
*/		
		}
	}

	/* label ports if applicable */
	if (portlabel) {
		l = cells->label;
		fprintf(psfile, "\n%% Port Labels\n");
		while (l != NULL) {
			fprintf(psfile, "%d %d moveto (%s) show\n", l->pos[0], l->pos[1], l->label);
			l = l->next;
		}
	}
	
	/* Finish page */
	fprintf(psfile,"\nshowpage\n");
	fclose(psfile);
}

int searchLayerMap(char *name)
{
	int i;

	if (name[0] == 0) return 0;
	for (i=0; i<numLayers; i++)
		if (strcmp(name,layers[i].name) == 0) return i;
	printf("undefined layer %s\n", name);
	return 0; /* if no layer is found with a matching name, return layer 0 as the null layer */
}

cellL *findCell(int cellNum)
{
	cellL *c = cells;
	while (1) {
		if (c == NULL || c->cellNum == cellNum) return c;
		c = c->next;
	}
}

boxL *copybox(boxL *g, float m[9])
{
	boxL *newbox;
	int dx, dy;

	if (g == NULL) {
		fprintf(stderr, "Null g pointer for copybox\n");
		exit(1);
	}
	newbox = (boxL*)malloc(sizeof(boxL));
	if (newbox == NULL) {
		fprintf(stderr, "copybox unable to allocate new box\n");
		exit(1);
	}
	newbox->layer = g->layer;
	newbox->next = NULL;
	transformBox(newbox->pos, g->pos, m);
	newbox->visible = g->visible;

	/* Update bounding box */
	dx = newbox->pos[0];
	dy = newbox->pos[1];
	if (newbox->pos[2] < cifBoundaries[0]) cifBoundaries[0] = newbox->pos[2];
	if (newbox->pos[3] < cifBoundaries[1]) cifBoundaries[1] = newbox->pos[3];
	if (newbox->pos[2]+dx > cifBoundaries[2]) cifBoundaries[2] = newbox->pos[2]+dx;
	if (newbox->pos[3]+dy > cifBoundaries[3]) cifBoundaries[3] = newbox->pos[3]+dy;
	return newbox;
}

polyL *copypoly(polyL *p, float m[9])
{
	polyL *newpoly;

	newpoly = (polyL*)malloc(sizeof(polyL));
	newpoly->layer = p->layer;
	newpoly->next = NULL;
	newpoly->numcoords = p->numcoords;
	transformPoly(newpoly, p, m);

	/* Update bounding box */
	/* *** not yet done for polygons */

	return newpoly;
}

cellL* insertCell(cellL *c, cellL *list)
{
	c->next = list;
	list = c;
	return list;
}

boxL* insertbox(boxL *g, boxL *list)
{
	g->next = list;
	list = g;
	return list;
}

polyL* insertpoly(polyL *p, polyL *list)
{
	p->next = list;
	list = p;
	return list;
}

labelL* insertlabel(labelL *p, labelL *list)
{
	p->next = list;
	list = p;
	return list;
}

cellInstL* insertInst(cellInstL *i, cellInstL *list)
{
	i->next = list;
	list = i;
	return list;
}

int isspacec(char c) {
   return (c == ',' || isspace(c));
}

int tokenize(char *line)
{
	int i=0, j=0, tok=0;
	int mode;

	if (isspacec(line[0])) mode = 0; 
	else mode=1;

	while (line[i] != 0 && line[i] != ';') {
		if (isspacec(line[i])) {
			if (mode == 1) {
				tokens[tok][j] = 0;
				mode = 0; j = 0;
				tok++;
			}
		}
		else {
			mode = 1;
			tokens[tok][j++] = line[i];
		}
		i++;
	}
	if (mode == 1) {
		tokens[tok][j] = 0;
		tok++;
	}
	return tok;
}


int linewrap(char *line)
{
	int i = 0;

	while (line[i] != 0) {
		if (line[i] == ';') return 0;
		i++;
	}
	return 1;
}

void matrixMul(float r[9], float a[9], float b[9])
{
	int i, j, k;
	float tmp[9];

	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			tmp[i*3+j] = 0;
			for (k=0; k<3; k++) {
				tmp[i*3+j] += a[i*3+k] * b[k*3+j];
			}
		}
	}
	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			r[i*3+j] = tmp[i*3+j];
		}
	}
}

void newIdentityMatrix(float m[9])
{
	int i,j;

	for (i=0; i<3; i++) {
		for (j=0; j<3; j++) {
			m[i*3+j] = (i==j);
		}
	}
}

void transformBox(int final[4], int initial[4], float m[9])
{
	int i,j;
	int pos[2];

	pos[0] = initial[2]+initial[0]/2;
	pos[1] = initial[3]+initial[1]/2;

	for (i=0; i < 2; i++) {
		final[i+2] = (int)m[6+i];
		for (j=0; j<2; j++) {
			final[i+2] += (int)m[i+j*3]*pos[j];
		}
	}

	if (m[1] == 0) { /* rotation */
		final[0] = initial[0];
		final[1] = initial[1];
	}
	else {
		final[0] = initial[1];
		final[1] = initial[0];
	}

	final[2] -= final[0]/2;
	final[3] -= final[1]/2;
}

void transformPoly(polyL *final, polyL *initial, float m[9])
{
	int p, i,j;
	
	for (p=0; p<initial->numcoords/2; p++) {
		for (i=0; i <2; i++) {
			final->coords[p*2+i] = (int)m[6+i];
			for (j=0; j<2; j++) {
				final->coords[p*2+i] += (int)m[i+j*3]*initial->coords[p*2+j];
			}
		}
	}
}

void printStatistics(void)
{
	printf("plotting statistics:\n");
	printf("  %d CIF Layers defined or transparencies implied in layer map\n", numLayers);
	printf("  %d cells defined with DF in CIF\n", numCells);
	printf("  %d instances used\n", numInstances);
	printf("  %d boxes generated\n", numBoxes);
	printf("  %d polygons generated\n", numPolys);
}

void convPDF(char *fname)
{
  char cmd[255];

  // issue system call to convert ps to PDF.  Requres ps2pdf installed.

  sprintf(cmd, "ps2pdf %s", fname);
  system(cmd);
}

void freeMem(void)
{
	int i;

	freeCell(cells);
	for (i=0; i<numLayers; i++) {
		freeBox(flattenedbox[i]);
		freePoly(flattenedpoly[i]);
		freeBoxQuadTree(boxQuadTrees[i]);
	}
}

void freeBoxQuadTree(boxquadtree *q)
{
	if (q != NULL) {
		freeBoxQuadTree(q->tl);
		freeBoxQuadTree(q->tr);
		freeBoxQuadTree(q->bl);
		freeBoxQuadTree(q->br);
		free(q->boxes);
		free(q);
	}
}

void freeCell(cellL *c)
{
	cellL *c2;

	while (c != NULL) {
		freeCellInst(c->inst);
		freeBox(c->box);
		freePoly(c->poly);
		freeLabel(c->label);
		c2 = c;
		c = c->next;
		free(c2);
	}
}

void freeBox(boxL *b) 
{
	boxL *b2;

	while (b != NULL) {
		b2 = b;
		b = b->next;
		free(b2);
	}
}

void freePoly(polyL *b) 
{
	polyL *b2;

	while (b != NULL) {
		b2 = b;
		b = b->next;
		free(b2);
	}
}

void freeLabel(labelL *b)
{
	labelL *b2;

	while (b != NULL) {
		b2 = b;
		b = b->next;
		free(b2);
	}
}

void freeCellInst(cellInstL *b)
{
	cellInstL *b2;

	while (b != NULL) {
		b2 = b;
		b = b->next;
		free(b2);
	}
}

void freeLine(lineL *b)
{
	lineL *b2;

	while (b != NULL) {
		b2 = b;
		b = b->next;
		free(b2);
	}
}


